{-|
Module      : Data.Graph.Clustering.ILouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

ILouvain: inductive Graph clustering with destructives updates

TODO FGL improvements:
- deg   :: Graph gr => gr a b -> Node -> Maybe Int
- match :: Graph gr => Node -> gr a b -> Maybe (Context, Graph)

-}

{-# LANGUAGE FlexibleContexts #-}

module Data.Graph.Clustering.ILouvain
  where

import Debug.SimpleReflect
import Data.Set (fromList, Set)
import qualified Data.Set as Set
import Data.Maybe (catMaybes, maybe)
import Data.List (zip, cycle, null, nub)
import Protolude hiding (empty, (&))
import Data.Graph.Inductive
import qualified Data.Graph.Clustering.HLouvain as H

data NodePath = AllNodes | DfsNodes

------------------------------------------------------------------------
------------------------------------------------------------------------
-- data MetaGraph = LabGraph (HyperGraph) | EdgeGraph (StreamGraph)
-- Metagraph (plural metagraphs) (mathematics) A representation
-- of a set of nodes and the morphisms relating them
-- https://en.wiktionary.org/wiki/metagraph
-- MetaGraph (x,y) where x is space (hyper structure of labels), y is
-- time like (stream structure)
-- (Below LabGraph definition)

-- | Why do we need more depth in the HyperGraph ? BUG to fix:
-- toNodes $ mv (mv (mv spoon [1] [2]) [2] [3]) [3,1] [] ==
-- [[1,2],[3,2],[4],[5],[6]] (Node 2 is in 2 different clusters,
-- not because it is a "graph non recouvrant" but because nodes are
-- forgeted when moving it)

type HyperGraph    a b   = Gr (Gr () a) b
type HyperContext  a b   = Context (Gr () a) b

type HyperGraph'   a b c = Gr (HyperGraph a b) c
type HyperContext' a b c = Context (HyperGraph a b) c

type GraphContexts  a b = Gr [Context () a] b

-- TODO: EdgeGraph (StreamGraph)
-- type StreamGraph  a b = Gr a (Gr () b)
------------------------------------------------------------------------

-- TODO
class IsHyperGraph gr where
  fromGraph_    :: Gr () a -> gr

  -- Flat is to HyperGraph what is empty to Graph
  -- Empty HyperGraph is flat
  -- but flat HyperGraph is not always empty
  -- Flat is empty for depth > 1 of Hypergraph
  -- type Depth = Int
  -- isFlat :: Depth -> Bool
  -- isFlat 0 = empty
  -- isFlat 1 = undefined
  isFlat_       :: gr -> Bool

  -- hnodes_ :: Node -> gr -> [Node]
  hnodes_       :: gr -> Node -> [Node]
  toNodes_      :: gr -> [[Node]]

  mv_           :: (Show gr, Graph gr, DynGraph gr)
                => gr -> [Node] -> [Node] -> gr
  -- hdeg_         :: a -> Node -> Maybe Int

  imodularity_  :: gr -> gr -> [Node] -> Double
  gmodularity_  :: gr -> gr -> Double

  -- Minimal Graphs for tests
  spoon_  :: gr
  karate_ :: gr
  -- TODO Random graphs generation

------------------------------------------------------------------------
instance IsHyperGraph (GraphContexts a a) where
  fromGraph_ = undefined

  isFlat_ :: GraphContexts a a -> Bool
  isFlat_ g = all null $ map snd (labNodes g)

  hnodes_ :: GraphContexts a a -> Node -> [Node]
  hnodes_ g n = case (match n g) of
    (Nothing        , _) -> []
    (Just (_,_,cs,_), _) -> map snd4 cs

  toNodes_ :: GraphContexts a a -> [[Node]]
  toNodes_ g = map (hnodes_ g) (nodes g)

  mv_ :: ( Show     (GraphContexts a a)
         , Graph    (GraphContexts a a)
         , DynGraph (GraphContexts a a)
         )
      => GraphContexts a a
      -> [Node] -> [Node]
      -> GraphContexts a a
  mv_ g [ ]    [ ] = g
  mv_ g [_]    [ ] = g
  mv_ g [ ]    [_] = g
  mv_ g [a,b] [ ]  = case match a g of
    (Nothing,  _) -> panic $ "mv: fst path does not exist: " <> show a
    (Just (p,n,cs,s), g1) -> case match b (foldr (&) empty cs) of
      (Just c, g2)  -> c & ( (p, n, gsel (const True) g2, s) & g1)
      (Nothing, _)               -> panic "mv: snd path does not exist"
  mv_ g [a,b] [c]  = mv (mv g [a,b] []) [b] [c]
  mv_ _ [_]   [_]  = panic "path too long"j


------------------------------------------------------------------------
type MaxIterations = Int
type MaxSize       = Int

type Start a = HyperGraph' a a a
type Trans a = HyperGraph' a a a
type Final a = HyperGraph' a a a
------------------------------------------------------------------------
iLouvain :: (Eq a, Show a)
         => MaxIterations
         -> MaxSize
         -> NodePath
         -> HyperGraph a a
         -> HyperGraph a a
iLouvain 0 s p g = g
iLouvain n s p g
  | length (toNodes g') <= s = g'
  | otherwise                = iLouvain' p g g'
    where
      g' = iLouvain (n-1) s p g

iLouvainH :: (Eq a, Show a)
         => MaxIterations
         -> MaxSize
         -> NodePath
         -> Start a
         -> Final a
iLouvainH 0 s p g = g
iLouvainH n s p g
  | length (toNodes' g') <= s = g'
  | otherwise                 = iLouvainH' p g g'
    where
      g' = iLouvainH (n-1) s p g
------------------------------------------------------------------------
iLouvain' :: (Eq a, Show a)
          => NodePath
          -> HyperGraph a a
          -> HyperGraph a a
          -> HyperGraph a a
iLouvain' p g g' =   -- trace (show ps :: Text)
  iLouvain'' g g' ps -- $ filter (\n -> elem n (nodes g)) ps
    where
      -- quick trick to filter path but path of HyperGraph can be different
      ps = if isFlat g'
              then case p of
                     AllNodes -> nodes g -- random path
                     DfsNodes -> path' g -- DFS path
              else sortOn (length . hnodes g') (nodes g')

iLouvainH' :: (Eq a, Show a)
          => NodePath
          -> Start a
          -> Trans a
          -> Final a
iLouvainH' p g g' =   -- trace (show ps :: Text)
  iLouvainH'' g g' ps -- $ filter (\n -> elem n (nodes g)) ps
    where
      -- quick trick to filter path but path of HyperGraph can be different
      ps = if isFlat' g'
              then case p of
                     AllNodes -> nodes g
                     DfsNodes -> path' g
              else sortOn (length . hnodes' g') (nodes g')

------------------------------------------------------------------------
iLouvain'' :: Show a
           => HyperGraph a a
           -> HyperGraph a a
           -> [Node]
           -> HyperGraph a a
iLouvain'' g g' ns  = foldl' (\g1 n -> trace (show (n, toNodes g1) :: Text) 
                             $ stepMax g g1 n
                             $ neighborhood g g1 n ) g' ns

iLouvainH'' :: Show a
           => Start a
           -> Trans a
           -> [Node]
           -> Final a
iLouvainH'' g g' ns  = foldl' (\g1 n -> trace (show (n, toNodes' g1) :: Text) 
                             $ stepMax' g g1 n
                             $ neighborhood' g g1 n ) g' ns
------------------------------------------------------------------------
-- | Neighborhood definition
-- TODO optim sort by jackard
-- TODO Tests
neighborhood :: HyperGraph a a -> HyperGraph a a -> Node -> [Node]
neighborhood g g' n = case isFlat g' of
  True  -> neighbors g' n
  False -> filter (\m -> m /= n
                      && (not . Set.null) (Set.intersection (fromList $ hnodes g m) (hood g g' n))
                  ) (nodes g')

neighborhood' :: HyperGraph' a a a -> HyperGraph' a a a -> Node -> [Node]
neighborhood' g g' n = case isFlat' g' of
  True  -> neighbors g' n
  False -> filter (\m -> m /= n
                      && (not . Set.null) (Set.intersection (fromList $ hnodes' g m) (hood' g g' n))
                  ) (nodes g')

-------------------------
hood :: HyperGraph a a -> HyperGraph a a -> Node -> Set Node
hood g g' n = H.exclusion (fromList ns) candidates
  where
    candidates = fromList $ concat $ map (neighbors g) ns
    ns         = hnodes g' n

hood' :: HyperGraph' a a a -> HyperGraph' a a a -> Node -> Set Node
hood' g g' n = H.exclusion (fromList ns) candidates
  where
    candidates = fromList $ concat $ map (neighbors g) ns
    ns         = hnodes' g' n

------------------------------------------------------------------------
stepMax :: Show a
     => HyperGraph a b
     -> HyperGraph a a
     -> Node
     -> [Node]
     -> HyperGraph a a
stepMax g g' x xs =
  maximumBy (\(g1) (g2) -> compare (gmodularity g g1) (gmodularity g g2))
                   $ scanl (\g'' (p,n) -> trace (show (p,n) :: Text) $
                                       if p == x
                                          then mv g'' [x]   [n]
                                          else mv g'' [p,x] [n]
                                          ) g'
                   $ zip (x:xs) xs
-------------------------------------------------------------------------
stepMax' :: Show a
     => Start a
     -> Trans a
     -> Node
     -> [Node]
     -> Final a
stepMax' g g' x xs =
  maximumBy (\(g1) (g2) -> compare (gmodularity' g g1) (gmodularity' g g2))
                   $ scanl (\g'' (p,n) -> trace (show (p,n) :: Text) $
                                       if p == x
                                          then mvH g'' [x]   [n]
                                          else mvH g'' [p,x] [n]
                                          ) g'
                   $ zip (x:xs) xs

------------------------------------------------------------------------
------------------------------------------------------------------------
{-
instance IsHyperGraph HyperGraph
  where
    toHyperGraph_ g = nmap (\_ -> empty) g
    isFlat_ = isFlat

    hnodes_ g n = case match n g of
      (Nothing          , _) -> []
      (Just (p, n, l, s), _) -> n : nodes l

    toNodes_ g = map (hnodes g) (nodes g)

    imodularity_ = imodularity
    gmodularity_ = gmodularity

    mv_ = mv
-}
------------------------------------------------------------------------
hnodes :: HyperGraph a b -> Node -> [Node]
hnodes g n = case match n g of
  (Nothing          , _) -> []
  (Just (p, n, l, s), _) -> n : nodes l

hnodes' :: HyperGraph' a b c -> Node -> [Node]
hnodes' g n = nub $ case match n g of
  (Nothing          , _) -> []
  (Just (p, n, l, s), _) -> n : (concat $ map (hnodes l) $ nodes l)

------------------------------------------------------------------------
toNodes :: HyperGraph a a -> [[Node]]
toNodes g = map (hnodes g) (nodes g)

toNodes' :: HyperGraph' a b c -> [[Node]]
toNodes' g = map (hnodes' g) (nodes g)

------------------------------------------------------------------------
isFlat :: HyperGraph a b -> Bool
isFlat g = all (isEmpty . snd) (labNodes g)

-- needs of flatten?
isFlat' :: HyperGraph' a b c -> Bool
isFlat' g = all (isFlat . snd) (labNodes g)

------------------------------------------------------------------------
------------------------------------------------------------------------
-- TODO go depth in HyperGraph (modularity at level/depth)
imodularity :: HyperGraph a b -> HyperGraph a a -> [Node] -> Double
imodularity g g' ns = -- trace ("imodul" :: Text) $
                   H.hmodularity g
                 $ fromList
                 $ concat
                 $ map (hnodes g') ns

imodularity' :: HyperGraph' a b c -> HyperGraph' a b c -> [Node] -> Double
imodularity' g g' ns =
                   H.hmodularity g
                 $ fromList
                 $ concat
                 $ map (hnodes' g') ns
------------------------------------------------------------------------
gmodularity :: HyperGraph a b -> HyperGraph a a -> Double
gmodularity g g' = sum $ map (\n -> imodularity g g' [n]) $ nodes g'

gmodularity' :: HyperGraph' a b c -> HyperGraph' a b c -> Double
gmodularity' g g' = sum $ map (\n -> imodularity' g g' [n]) $ nodes g'

------------------------------------------------------------------------
------------------------------------------------------------------------
toHyperGraph :: Gr () Double -> HyperGraph Double Double
toHyperGraph g = nmap (\_ -> empty) g

toHyperGraph' :: Gr () Double -> HyperGraph' Double Double Double
toHyperGraph' g = nmap (\_ -> empty) g
------------------------------------------------------------------------
------------------------------------------------------------------------
-- Spoon Graph
--    1
--   / \
--  2   3
--   \ /
--    4
--    |
--    5
spoon :: HyperGraph Double Double
spoon = mkGraph ns es
  where
    ns :: [LNode (Gr () Double)]
    ns = zip [1..6] (cycle [empty])

    es :: [LEdge Double]
    es = [ (1, 2, 1.0)
         , (1, 3, 1.0)
         , (2, 4, 1.0)
         , (3, 4, 1.0)
         , (4, 5, 1.0)
         ]

singleton :: Node -> HyperGraph Double Double
singleton n = mkGraph [(n,empty)] []

spoon' :: HyperGraph' Double Double Double
spoon' = mkGraph ns es
  where
    ns :: [LNode (HyperGraph Double Double)]
    ns = zip [1..] (map singleton [1..6]) --  (cycle [emptyHyperGraph])

    es :: [LEdge Double]
    es = [ (1, 2, 1.0)
         , (1, 3, 1.0)
         , (2, 4, 1.0)
         , (3, 4, 1.0)
         , (4, 5, 1.0)
         ]

-- | Needed function
-- mv: a Node elsewhere in the HyperGraph

-- Move Properties:
-- mv Node into another Node and remove it again should be equal to
-- identity

-- property test:
-- mv (mv g [a] [b]) [b,a] [] = identity
-- mv g [a] [] == g
-- mv g [a,b] [] /= g
-- mv (mv spoon [1] [4]) [4,1] [] = identity

-- Move target type
-- let's start simple: path lenght <= 2 max
mv :: Show a => HyperGraph a a
   -> [Node] -> [Node]
   -> HyperGraph a a
mv g [ ] [ ] = g
mv g [_] [ ] = g
mv g [ ] [_] = g
mv g [a] [b] = case a == b of
  True  -> panic "mv, impossible: from node == dest node"
  False -> mv' g a b
mv g [a,b] [ ] = case match a g of
  (Nothing,  _) -> panic $ "mv: fst path does not exist: " <> show a
  (Just (p, n, l, s), g1) -> case match b l of
    (Just (p',n',l',s'), g2)  -> (p', n', g2, s')
                               & ( (p , n , delNode b l , s )
                               &  g1 )
    (Nothing, _) -> panic "mv: snd path does not exist"
mv g [a,b] [c] = mv (mv g [a,b] []) [b] [c]
mv g _ _  = panic "mv:todo long path"

                  ----------------------------

-- | Start simple (without path)
mv' :: Show a => HyperGraph a a
    -> Node -> Node
    -> HyperGraph a a
mv' g n1 n2 = -- trace (show (c1,c2) :: Text) $
  (mvMContext c1 c2) & g2
    where
      (c1, g1) = match n1 g
      (c2, g2) = match n2 g1

mvMContext :: Maybe (HyperContext a a)
           -> Maybe (HyperContext a a)
           -> HyperContext a a
mvMContext (Just (a1 ,n ,l ,a2 ))
           (Just (a1',n',l',a2')) = (a1',n', merge l' (c & l),a2')
            where
              c = (a1, n, (), a2)
mvMContext _ (Just _) = panic "mvMContext: First Node  does not exist"
mvMContext (Just _) _ = panic "mvMContext: Snd   Node  does not exist"
mvMContext _ _        = panic "mvMContext: Both  Nodes do   not exist"


------------------
-- Test

test_mv_mv = g0 == foldl' (\g (n1,n2) -> trace (prettify g) $ mv g n1 n2) g0 ns
  where
    g0 = spoon
    ns = [ ([2], [1])
         , ([3], [1])
         , ([1,2], [])
         --, ([1,3], [])
         ]

------------------

mvH :: Show a
    => HyperGraph' a a a
    -> [Node] -> [Node]
    -> HyperGraph' a a a
mvH g [ ] [ ] = g
mvH g [_] [ ] = g
mvH g [ ] [_] = g
mvH g [a] [b] = case a == b of
  True  -> panic "mv, impossible: from node == dest node"
  False -> mvH' g a b
mvH g [a,b] [ ] = case match a g of
  (Nothing,  _) -> panic $ "mv: fst path does not exist: " <> show a
  (Just (p, n, l, s), g1) -> case match b l of
    (Just (p',n',l',s'), g2)  -> (p', n', g2, s')
                               & ((p , n , delNode b l , s )
                               & g1)
    (Nothing, _) -> panic "mv: snd path does not exist"
mvH g [a,b] [c] = mvH (mvH g [a,b] []) [b] [c]
mvH g _ _  = panic "mv:todo long path"


mvH' :: Show a => HyperGraph' a a a
    -> Node -> Node
    -> HyperGraph' a a a
mvH' g n1 n2 = -- trace (show (c1,c2) :: Text) $
  (mvMContext' c1 c2) & g2
    where
      (c1, g1) = match n1 g
      (c2, g2) = match n2 g1
--{-
mvMContext' :: Show a
            => Maybe (HyperContext' a a a)
            -> Maybe (HyperContext' a a a)
            -> HyperContext' a a a
mvMContext' (Just (a1 ,n ,l ,a2 ))
            -- (Just (a1',n',l',a2')) = (a1',n',  insNode (n,l) (c & l'), a2')
            (Just (a1',n',l',a2')) = (a1',n', merge l' (c & l), a2')
            where
              c = (a1, n, empty , a2)
mvMContext' _ (Just _) = panic "mvMContext': First Node  does not exist"
mvMContext' (Just _) _ = panic "mvMContext': Snd   Node  does not exist"
mvMContext' _ _        = panic "mvMContext': Both  Nodes do   not exist"
--}

------------------------------------------------------------------------
test_mv :: (Ord a, Show a) => HyperGraph a a -> Node -> Node -> Bool
test_mv g a b = (mv (mv g [a] [b]) [b,a] []) == g
------------------------------------------------------------------------
-- Paths in the Graph to be tested
-- Directed graph strategy
path_dir :: Graph gr => gr a b -> [[Node]]
path_dir g = map (\xs -> dfs xs g ) (components g)

-- UnDirected graph strategy
path :: (DynGraph gr, Eq b) => gr a b -> [[Node]]
path g' = map sortNodes cs
  where
    sortNodes ns = case head $ sortOn (Down . (deg g)) ns of
      Nothing -> []
      Just  n -> dfs [n] g  -- dfs for glustering, bfs for klustering
    cs = components g
    g  = undir g'

path' :: (DynGraph gr, Eq b) => gr a b -> [Node]
path' = maybe [] identity  . head . path

------------------------------------------------------------------------
-- FGL specific functions
merge :: (Graph gr, DynGraph gr)
      => gr a b -> gr a b -> gr a b
merge = ufold (&)
------------------------------------------------------------------------
-- Tuple 4 functions
-- since Data.Tuple.Extra has Tuple3 sugar only
fst4 :: (a,b,c,d) -> a
fst4    (a,_,_,_)  = a

snd4 :: (a,b,c,d) -> b
snd4    (_,b,_,_)  = b

thd4 :: (a,b,c,d) -> c
thd4    (_,_,c,_)  = c

fth4 :: (a,b,c,d) -> d
fth4    (_,_,_,d)  = d
------------------------------------------------------------------------
------------------------------------------------------------------------
-- TO DELETE

step :: Show a
     => HyperGraph a b
     -> HyperGraph a a
     -> Node
     -> [Node]
     -> HyperGraph a a
step g g' n ns = -- trace ("step'" :: Text) $
  foldl' (\g1 n' -> case match n g1 of
           (Nothing, _) -> g1
           (Just _, _ ) -> step' g g1 n n'
          ) g' ns

step' :: Show a
     => HyperGraph a b
     -> HyperGraph a a
     -> Node
     -> Node
     -> HyperGraph a a
step' g g' n1 n2 = -- trace ("step " <> show n1 <> " " <> show n2 :: Text) $
  if s2 > 0 && s2 >= s1
  -- if s2 >= s1
    then -- trace ("step:mv " <> show n1 <> " " <> show n2 :: Text) $
      mv g' [n1] [n2]
    else -- trace ("step:else" :: Text) $
      g'
  where
    s1 = -- trace ("step:mod1" :: Text) $
      imodularity g g' [n1]
    s2 = -- trace ("step:mod2" :: Text) $
      imodularity g g' [n1,n2]
